@@ -19,6 +19,8 @@
 #include "config.h"
 
 #include "utils.h"
+#include <limits.h>
+#include <stdint.h>
 #include <sys/syscall.h>
 #include <sys/socket.h>
 #ifdef HAVE_SELINUX
@@ -33,10 +35,16 @@
 #endif
 
 __attribute__((format(printf, 1, 0))) static void
-warnv (const char *format, va_list args)
+warnv (const char *format,
+       va_list args,
+       const char *detail)
 {
   fprintf (stderr, "bwrap: ");
   vfprintf (stderr, format, args);
+
+  if (detail != NULL)
+    fprintf (stderr, ": %s", detail);
+
   fprintf (stderr, "\n");
 }
 
@@ -46,7 +54,7 @@ warn (const char *format, ...)
   va_list args;
 
   va_start (args, format);
-  warnv (format, args);
+  warnv (format, args, NULL);
   va_end (args);
 }
 
@@ -58,13 +66,24 @@ die_with_error (const char *format, ...)
 
   errsv = errno;
 
-  fprintf (stderr, "bwrap: ");
-
   va_start (args, format);
-  vfprintf (stderr, format, args);
+  warnv (format, args, strerror (errsv));
   va_end (args);
 
-  fprintf (stderr, ": %s\n", strerror (errsv));
+  exit (1);
+}
+
+void
+die_with_mount_error (const char *format, ...)
+{
+  va_list args;
+  int errsv;
+
+  errsv = errno;
+
+  va_start (args, format);
+  warnv (format, args, mount_strerror (errsv));
+  va_end (args);
 
   exit (1);
 }
@@ -75,14 +94,14 @@ die (const char *format, ...)
   va_list args;
 
   va_start (args, format);
-  warnv (format, args);
+  warnv (format, args, NULL);
   va_end (args);
 
   exit (1);
 }
 
 void
-die_unless_label_valid (const char *label)
+die_unless_label_valid (UNUSED const char *label)
 {
 #ifdef HAVE_SELINUX
   if (is_selinux_enabled () == 1)
@@ -126,9 +145,9 @@ xmalloc (size_t size)
 }
 
 void *
-xcalloc (size_t size)
+xcalloc (size_t nmemb, size_t size)
 {
-  void *res = calloc (1, size);
+  void *res = calloc (nmemb, size);
 
   if (res == NULL)
     die_oom ();
@@ -138,9 +157,13 @@ xcalloc (size_t size)
 void *
 xrealloc (void *ptr, size_t size)
 {
-  void *res = realloc (ptr, size);
+  void *res;
 
-  if (size != 0 && res == NULL)
+  assert (size != 0);
+
+  res = realloc (ptr, size);
+
+  if (res == NULL)
     die_oom ();
   return res;
 }
@@ -568,7 +591,6 @@ load_file_data (int     fd,
   ssize_t data_read;
   ssize_t data_len;
   ssize_t res;
-  int errsv;
 
   data_read = 0;
   data_len = 4080;
@@ -578,6 +600,12 @@ load_file_data (int     fd,
     {
       if (data_len == data_read + 1)
         {
+          if (data_len > SSIZE_MAX / 2)
+            {
+              errno = EFBIG;
+              return NULL;
+            }
+
           data_len *= 2;
           data = xrealloc (data, data_len);
         }
@@ -587,12 +615,7 @@ load_file_data (int     fd,
       while (res < 0 && errno == EINTR);
 
       if (res < 0)
-        {
-          errsv = errno;
-          close (fd);
-          errno = errsv;
-          return NULL;
-        }
+        return NULL;
 
       data_read += res;
     }
@@ -672,7 +695,7 @@ ensure_dir (const char *path,
 /* Sets errno on error (!= 0) */
 int
 mkdir_with_parents (const char *pathname,
-                    int         mode,
+                    mode_t      mode,
                     bool        create_last)
 {
   cleanup_free char *fn = NULL;
@@ -809,13 +832,15 @@ readlink_malloc (const char *pathname)
 
   do
     {
+      if (size > SIZE_MAX / 2)
+        die ("Symbolic link target pathname too long");
       size *= 2;
       value = xrealloc (value, size);
       n = readlink (pathname, value, size - 1);
       if (n < 0)
         return NULL;
     }
-  while (size - 2 < n);
+  while (size - 2 < (size_t)n);
 
   value[n] = 0;
   return steal_pointer (&value);
@@ -829,6 +854,14 @@ get_oldroot_path (const char *path)
   return strconcat ("/oldroot/", path);
 }
 
+char *
+get_newroot_path (const char *path)
+{
+  while (*path == '/')
+    path++;
+  return strconcat ("/newroot/", path);
+}
+
 int
 raw_clone (unsigned long flags,
            void         *child_stack)
@@ -854,7 +887,7 @@ pivot_root (const char * new_root, const char * put_old)
 }
 
 char *
-label_mount (const char *opt, const char *mount_label)
+label_mount (const char *opt, UNUSED const char *mount_label)
 {
 #ifdef HAVE_SELINUX
   if (mount_label)
@@ -871,7 +904,7 @@ label_mount (const char *opt, const char *mount_label)
 }
 
 int
-label_create_file (const char *file_label)
+label_create_file (UNUSED const char *file_label)
 {
 #ifdef HAVE_SELINUX
   if (is_selinux_enabled () > 0 && file_label)
@@ -881,7 +914,7 @@ label_create_file (const char *file_label)
 }
 
 int
-label_exec (const char *exec_label)
+label_exec (UNUSED const char *exec_label)
 {
 #ifdef HAVE_SELINUX
   if (is_selinux_enabled () > 0 && exec_label)
@@ -889,3 +922,24 @@ label_exec (const char *exec_label)
 #endif
   return 0;
 }
+
+/*
+ * Like strerror(), but specialized for a failed mount(2) call.
+ */
+const char *
+mount_strerror (int errsv)
+{
+  switch (errsv)
+    {
+      case ENOSPC:
+        /* "No space left on device" misleads users into thinking there
+         * is some sort of disk-space problem, but mount(2) uses that
+         * errno value to mean something more like "limit exceeded". */
+        return ("Limit exceeded (ENOSPC). "
+                "(Hint: Check that /proc/sys/fs/mount-max is sufficient, "
+                "typically 100000)");
+
+      default:
+        return strerror (errsv);
+    }
+}
